/**
 * @license
 * Copyright Google LLC All Rights Reserved.
 *
 * Use of this source code is governed by an MIT-style license that can be
 * found in the LICENSE file at https://angular.io/license
 */
import { ERROR_DETAILS_PAGE_BASE_URL } from "./error_details_base_url";
/**
 * The list of error codes used in runtime code of the `core` package.
 * Reserved error code range: 100-999.
 *
 * Note: the minus sign denotes the fact that a particular code has a detailed guide on
 * angular.io. This extra annotation is needed to avoid introducing a separate set to store
 * error codes which have guides, which might leak into runtime code.
 *
 * Full list of available error guides can be found at https://angular.io/errors.
 */

export let RuntimeErrorCode;
/**
 * Class that represents a runtime error.
 * Formats and outputs the error message in a consistent way.
 *
 * Example:
 * ```
 *  throw new RuntimeError(
 *    RuntimeErrorCode.INJECTOR_ALREADY_DESTROYED,
 *    ngDevMode && 'Injector has already been destroyed.');
 * ```
 *
 * Note: the `message` argument contains a descriptive error message as a string in development
 * mode (when the `ngDevMode` is defined). In production mode (after tree-shaking pass), the
 * `message` argument becomes `false`, thus we account for it in the typings and the runtime logic.
 */

(function (RuntimeErrorCode) {
  RuntimeErrorCode[RuntimeErrorCode["EXPRESSION_CHANGED_AFTER_CHECKED"] = -100] = "EXPRESSION_CHANGED_AFTER_CHECKED";
  RuntimeErrorCode[RuntimeErrorCode["RECURSIVE_APPLICATION_REF_TICK"] = 101] = "RECURSIVE_APPLICATION_REF_TICK";
  RuntimeErrorCode[RuntimeErrorCode["CYCLIC_DI_DEPENDENCY"] = -200] = "CYCLIC_DI_DEPENDENCY";
  RuntimeErrorCode[RuntimeErrorCode["PROVIDER_NOT_FOUND"] = -201] = "PROVIDER_NOT_FOUND";
  RuntimeErrorCode[RuntimeErrorCode["INVALID_FACTORY_DEPENDENCY"] = 202] = "INVALID_FACTORY_DEPENDENCY";
  RuntimeErrorCode[RuntimeErrorCode["MISSING_INJECTION_CONTEXT"] = 203] = "MISSING_INJECTION_CONTEXT";
  RuntimeErrorCode[RuntimeErrorCode["INVALID_INJECTION_TOKEN"] = 204] = "INVALID_INJECTION_TOKEN";
  RuntimeErrorCode[RuntimeErrorCode["INJECTOR_ALREADY_DESTROYED"] = 205] = "INJECTOR_ALREADY_DESTROYED";
  RuntimeErrorCode[RuntimeErrorCode["MULTIPLE_COMPONENTS_MATCH"] = -300] = "MULTIPLE_COMPONENTS_MATCH";
  RuntimeErrorCode[RuntimeErrorCode["EXPORT_NOT_FOUND"] = -301] = "EXPORT_NOT_FOUND";
  RuntimeErrorCode[RuntimeErrorCode["PIPE_NOT_FOUND"] = -302] = "PIPE_NOT_FOUND";
  RuntimeErrorCode[RuntimeErrorCode["UNKNOWN_BINDING"] = 303] = "UNKNOWN_BINDING";
  RuntimeErrorCode[RuntimeErrorCode["UNKNOWN_ELEMENT"] = 304] = "UNKNOWN_ELEMENT";
  RuntimeErrorCode[RuntimeErrorCode["TEMPLATE_STRUCTURE_ERROR"] = 305] = "TEMPLATE_STRUCTURE_ERROR";
  RuntimeErrorCode[RuntimeErrorCode["INVALID_EVENT_BINDING"] = 306] = "INVALID_EVENT_BINDING";
  RuntimeErrorCode[RuntimeErrorCode["MULTIPLE_PLATFORMS"] = 400] = "MULTIPLE_PLATFORMS";
  RuntimeErrorCode[RuntimeErrorCode["PLATFORM_NOT_FOUND"] = 401] = "PLATFORM_NOT_FOUND";
  RuntimeErrorCode[RuntimeErrorCode["ERROR_HANDLER_NOT_FOUND"] = 402] = "ERROR_HANDLER_NOT_FOUND";
  RuntimeErrorCode[RuntimeErrorCode["BOOTSTRAP_COMPONENTS_NOT_FOUND"] = 403] = "BOOTSTRAP_COMPONENTS_NOT_FOUND";
  RuntimeErrorCode[RuntimeErrorCode["ALREADY_DESTROYED_PLATFORM"] = 404] = "ALREADY_DESTROYED_PLATFORM";
  RuntimeErrorCode[RuntimeErrorCode["ASYNC_INITIALIZERS_STILL_RUNNING"] = 405] = "ASYNC_INITIALIZERS_STILL_RUNNING";
  RuntimeErrorCode[RuntimeErrorCode["INVALID_I18N_STRUCTURE"] = 700] = "INVALID_I18N_STRUCTURE";
  RuntimeErrorCode[RuntimeErrorCode["INVALID_DIFFER_INPUT"] = 900] = "INVALID_DIFFER_INPUT";
  RuntimeErrorCode[RuntimeErrorCode["NO_SUPPORTING_DIFFER_FACTORY"] = 901] = "NO_SUPPORTING_DIFFER_FACTORY";
  RuntimeErrorCode[RuntimeErrorCode["VIEW_ALREADY_ATTACHED"] = 902] = "VIEW_ALREADY_ATTACHED";
  RuntimeErrorCode[RuntimeErrorCode["INVALID_INHERITANCE"] = 903] = "INVALID_INHERITANCE";
  RuntimeErrorCode[RuntimeErrorCode["UNSAFE_VALUE_IN_RESOURCE_URL"] = 904] = "UNSAFE_VALUE_IN_RESOURCE_URL";
  RuntimeErrorCode[RuntimeErrorCode["UNSAFE_VALUE_IN_SCRIPT"] = 905] = "UNSAFE_VALUE_IN_SCRIPT";
})(RuntimeErrorCode || (RuntimeErrorCode = {}));

export class RuntimeError extends Error {
  constructor(code, message) {
    super(formatRuntimeError(code, message));
    this.code = code;
  }

}
/**
 * Called to format a runtime error.
 * See additional info on the `message` argument type in the `RuntimeError` class description.
 */

export function formatRuntimeError(code, message) {
  // Error code might be a negative number, which is a special marker that instructs the logic to
  // generate a link to the error details page on angular.io.
  const fullCode = `NG0${Math.abs(code)}`;
  let errorMessage = `${fullCode}${message ? ": " + message : ""}`;

  if (code < 0) {
    errorMessage = `${errorMessage}. Find more at ${ERROR_DETAILS_PAGE_BASE_URL}/${fullCode}`;
  }

  return errorMessage;
}